/****************************************/
/* MES/Copyleft Yukio Mituiwa,2005	*/
/*					*/
/*  2006/6/16 first release		*/
/*					*/
/****************************************/
#include	<macro.h>
#include	<module.h>
#include	<string.h>

#ifdef _KERNEL_
  #include	"mes.h"
  #include	"conf.h"
  #define device_main(arg1,arg2) init_ata(void)
  extern BUSConf hdd_config[];
#endif
#include	"ata.h"

typedef struct {
	volatile unsigned short	*base0, *base1;
	unsigned int		offset[2];
	unsigned int		size[2];
	unsigned char		port, bit;
} ATAInfo;
#define ATANUM	2
#define WAITMS	1000
#define RETRY_NUM	4
static ATAInfo	ata[ATANUM];

static int busy_wait(int minor) {
	int	cnt;

	cnt = 0;
	while(ATA_SR & SR_BSY) {
		if(cnt++ > WAITMS) return -1;
		task_sw_trap();
	}
	return 0;
}

static int device_selection(int minor) {
	int	cnt;

	if(busy_wait(minor) == -1) return -1;
	ATA_DHR = (minor & 1) ? DHR_DEV : 0;
	task_sw_trap();
	cnt = 0;
	while((ATA_SR &(SR_BSY | SR_DRDY)) != SR_DRDY){
		if(cnt++ > WAITMS) return -1;
		task_sw_trap();
	}
	return 0;
}

/*
static int non_data_command(int minor, unsigned short command, int count, int lba) {
	if(device_selection(minor) == -1) return -1;
	ATA_SCR = count;
	ATA_SNR = lba & 0xff;
	ATA_CYLL = (lba >> 8) & 0xff;
	ATA_CYLH = (lba >> 16) & 0xff;
	ATA_DHR |= DHR_LBA | ((lba >> 24) & 0xf);
	ATA_CR = command;
	if(busy_wait(minor) == -1) return -1;
	if(ATA_SR & (SR_DF | SR_ERR)) return -1;
	return 0;
}

static int ata_identify(int minor, short *data) {
	int i, s;

	if(device_selection(minor) == -1) return -1;
	ATA_CR = IdentifyDevice;
	if(busy_wait(minor) == -1) return -1;
	if(ATA_SR & (SR_DF | SR_ERR)) return -1;
	s = ATA_SR;
	if(s & SR_DREQ) {
		for(i = 0;i < 256;i++) data[i] =  ATA_DR;
	}
	return 0;
}
*/

static int data_in_command(int minor, unsigned short command, char *data, int lba, int count) {
	int	i, s;
	short 	value;

	if(device_selection(minor) == -1) return -1;
	ATA_SCR = count;
	ATA_SNR = lba & 0xff;
	ATA_CYLL = (lba >> 8) & 0xff;
	ATA_CYLH = (lba >> 16) & 0xff;
	ATA_DHR |= DHR_LBA | ((lba >> 24) & 0xf);
	ATA_CR = command;
	if(busy_wait(minor) == -1) return -1;
	if(ATA_SR & (SR_DF | SR_ERR)) return -1;
	s = ATA_SR;
	if(s & SR_DREQ) {
		for(i = 0;i < 512 * count;) {
			value =  ATA_DR;
			data[i++] = value & 0xff;
			data[i++] = (value >> 8) & 0xff;
		}
		return 0;
	}
	return -1;
}

static int data_out_command(int minor, unsigned short command, char *data, int lba, int count) {
	int	i, s;
	short 	value;

	if(device_selection(minor) == -1) return -1;
	ATA_SCR = count;
	ATA_SNR = lba & 0xff;
	ATA_CYLL = (lba >> 8) & 0xff;
	ATA_CYLH = (lba >> 16) & 0xff;
	ATA_DHR |= DHR_LBA | ((lba >> 24) & 0xf);
	ATA_CR = command;
	if(busy_wait(minor) == -1) return -1;
	if(ATA_SR & (SR_DF | SR_ERR)) return -1;
	s = ATA_SR;
	if(s & SR_DREQ) {
		for(i = 0;i < 512 * count;) {
			value = data[i++] & 0xff;
			value |= (data[i++] << 8) & 0xff00;
			ATA_DR = value;
		}
		if(ATA_SR & (SR_DF | SR_ERR)) return -1;
		return 0;
	}
	return -1;
}

static int ata_driveinfo(int minor, char *name, int *chs) {
	unsigned short	pio;
	unsigned short	data;
	int		i, s;
	unsigned short	*hddname;

	if(device_selection(minor) == -1) return -1;
	hddname = (unsigned short*)name;
	ATA_CR = IdentifyDevice;
	if(busy_wait(minor) == -1) return -1;
	if(ATA_SR & (SR_DF | SR_ERR)) return -1;
	s = ATA_ASR;
	pio = 0;
	if(s & SR_DREQ) {
		for(i = 0;i < 256;i++) {
			data = ATA_DR;
			if(i >= 27 && i < 47) hddname[i - 27] = data;
			switch(i) {
			case 54:
				chs[1] = data;
				break;
			case 55:
				chs[0] = data;
				break;
			case 56:
				chs[2] = data;
				break;
			case 64:
				pio = data & 0xff;
				break;
			}
		}
		ata[minor >> 1].size[minor & 1] = chs[0] * chs[1] * chs[2];
		return pio;
	}
	return -1;
}
/*
static int get_hddsize(int minor) {
	return ata[minor >> 1].size[minor & 1];
}
*/
static int ata_reset(int minor) {
	int	n;

	for(n = 0;n < 200;n++) {
		ATA_DCR = DCR_SRST;
		task_sw_trap();
		ATA_DCR = 0;
		task_sw_trap();
		if(busy_wait(minor) == 0) return 0;
	}
	return -1;
}

static int open_ata(int minor, int option) {
	char		name[40];
	int		chs[3], pio;
	IOREG		*datareg;

	if(minor > (ATANUM * 2)) return -1;
	if(ata[minor].base1 == 0) return -1;
#ifdef _KERNEL_
	set_buswidth((int)ata[minor >> 1].base0, 16, 64);
#endif
#if defined(__H8300H__) || defined(__sh__)
	ddr_set(ata[minor >> 1].port, 1 << ata[minor >> 1].bit);
	datareg = get_dr(ata[minor >> 1].port);
	*datareg |= 1 << ata[minor >> 1].bit;
#endif
	if(ata_reset(minor) == -1) return -1;
	ATA_DCR = DCR_nIEN;
	ata[minor >> 1].size[minor & 1] = 0;
	pio = ata_driveinfo(minor, name, chs);
	if(pio == -1) return -1;
	ata[minor >> 1].offset[minor & 1] = 0;
	return 0;
}

static int close_ata(int minor) {
	if(minor > (ATANUM * 2)) return -1;
	ata[minor >> 1].size[minor & 1] = 0;
	return 0;
}

static int write_ata(int minor, char *buffer, int size) {
	int	ret, try;

	if(minor > (ATANUM * 2)) return -1;
//	if(ata[minor >> 1].size[minor & 1] == 0) return -1;
	for(try = 0;try < RETRY_NUM;try++) {
		ret = data_out_command(minor, WriteVerify, buffer, ata[minor >> 1].offset[minor & 1], size / SECSIZ);
		if(ret != -1) break;
		ret = -1;
	}
	return (ret == -1) ? -1 : size;
}

static int read_ata(int minor, char *buffer, int size) {
	int	ret, try;

	if(minor > (ATANUM * 2)) return -1;
//	if(ata[minor >> 1].size[minor & 1] == 0) return -1;
	for(try = 0;try < RETRY_NUM;try++) {
		ret = data_in_command(minor, ReadSector, buffer, ata[minor >> 1].offset[minor & 1], size / SECSIZ);
		if(ret != -1) break;
		ret = -1;
	}
	return (ret == -1) ? -1 : size;
}

static int seek_ata(int minor, int pos) {
	if(minor > (ATANUM * 2)) return -1;
	ata[minor >> 1].offset[minor & 1] = pos / SECSIZ;
	return 0;
}

static int ioctl_ata(int minor, int data, int op) {
	int	ret;

	if(minor > (ATANUM * 2)) return -1;
	if(ata[minor >> 1].size[minor & 1] == 0) return -1;
	ret = 0;
	switch(op) {
	case DEV_INFO:
		break;
	default:
		ret = -1;
	}
	return ret;
}

static Functions func;

int device_main(int argc, char **argv) {
	int		minor, base0[ATANUM], base1[ATANUM];

#ifndef _KERNEL_
	int		n, p; /* ----- Append 2010/01/26 A.Ohnishi ----- */
	for(minor = 0;minor < ATANUM;minor++) {
		ata[minor].port = 0xff;
		ata[minor].bit = 0xff;
		base0[minor] = 0;
		base1[minor] = 0;
	}
	minor = 0;
	for(n = 1;n < argc;n++) {
		if(memcmp(argv[n], "hdd", 3) == 0) {
			minor = stoi(&argv[n][3]) >> 1;
			if(minor >= ATANUM) minor = 0;
		} else if(memcmp(argv[n], "b=", 2) == 0) {
			base0[minor] = stoh(&argv[n][2]);
		} else if(memcmp(argv[n], "s=", 2) == 0) {
			base1[minor] = stoh(&argv[n][2]);
		} else if(memcmp(argv[n], "port=", 5) == 0) {
			p = stoh(&argv[n][5]);
			ata[minor].port = (p >> 4) & 0x0f;
			ata[minor].bit = p & 0x0f;
		}
	}
#endif
	for(minor = 0;minor < ATANUM;minor++) {
#ifdef _KERNEL_
		base0[minor] = hdd_config[minor << 1].memaddr;
		base1[minor] = hdd_config[minor << 1].memsize;
		ata[minor].port = hdd_config[minor << 1].res_port;
		ata[minor].bit = hdd_config[minor << 1].res_bit;
#endif
		ata[minor].base0 = (volatile unsigned short*)base0[minor];
		ata[minor].base1 = (volatile unsigned short*)(base0[minor] + base1[minor]);
		if(ata[minor].base1 == 0) continue;
		ata[minor].size[0] = 0;
		ata[minor].size[1] = 0;
		ata[minor].offset[0] = 0;
		ata[minor].offset[1] = 0;
#ifndef _KERNEL_
		set_buswidth(base0[minor], 16, 64);
#endif
	}
	func.open_dev = open_ata;
	func.close_dev = close_ata;
	func.write_dev = write_ata;
	func.read_dev = read_ata;
	func.seek_dev = seek_ata;
	func.ioctl_dev = ioctl_ata;
	func.poll_dev = 0;
	strcpy(&(func.name[0]), "hdd");
	device_return(&func);
}
