/****************************************/
/* MES/Copyleft Yukio Mituiwa,2005	*/
/*					*/
/*  2006/6/16 first release		*/
/*					*/
/****************************************/
#include <module.h>
#include <string.h>

#ifdef _KERNEL_
  #include "mes.h"
  #include "conf.h"
  #define device_main(arg1,arg2) init_lcd(void)
  extern PIOConf	lcd_config[];
#endif

#define LCDNUM	4
typedef struct {
	int	width, height;
	int	cur;
	int	port;
	int	bits[6];
#if defined(__H8300H__) || defined(__sh__)
	IOREG	*dr;
#endif
} LCDInfo;

static LCDInfo lcd[LCDNUM];
static void wait_(int);
static void lcd_out8(int, unsigned char);
static void lcd_out4(int, int, unsigned char);
static void lcd_clear(int);
static void lcd_putch(int, int);

#define LCD_D4	lcd[minor].bits[0]
#define LCD_D5	lcd[minor].bits[1]
#define LCD_D6	lcd[minor].bits[2]
#define LCD_D7	lcd[minor].bits[3]
#define LCD_RS	lcd[minor].bits[4]
#define LCD_E	lcd[minor].bits[5]

static void wait_(int times) {
	volatile int w;
	for(w = 0;w < times * WAITCONST;w++);
}

static void lcd_out8(int minor, unsigned char dat) {
	int	d;
	
#if defined(__arm__)
	PIO_CODR = LCD_RS;
	PIO_SODR = LCD_E;
	PIO_CODR = LCD_D4 + LCD_D5 + LCD_D6 + LCD_D7;
	PIO_SODR = (dat & 0x08) ? LCD_D7 : 0;
	PIO_SODR = (dat & 0x04) ? LCD_D6 : 0;
	PIO_SODR = (dat & 0x02) ? LCD_D5 : 0;
	PIO_SODR = (dat & 0x01) ? LCD_D4 : 0;
	wait_(83);	// wait_ 83us
	PIO_CODR = LCD_E;
#endif
#if defined(__H8300H__) || defined(__sh__)
	d = LCD_E;
	d |= (dat & 0x08) ? LCD_D7 : 0;
	d |= (dat & 0x04) ? LCD_D6 : 0;
	d |= (dat & 0x02) ? LCD_D5 : 0;
	d |= (dat & 0x01) ? LCD_D4 : 0;
	*(lcd[minor].dr) = d;
	wait_(83);	// wait_ 83us
	d &= ~LCD_E;
	*(lcd[minor].dr) = d;
#endif
	wait_(4000);	// wait_ 4ms
}

static void lcd_out4(int minor, int rs, unsigned char dat) {
	int d;
	
#if defined(__arm__)
	if(rs == 0) {
		PIO_CODR = LCD_RS;
	} else {
		PIO_SODR = LCD_RS;
	}
	PIO_SODR = LCD_E;
	PIO_CODR = LCD_D4 + LCD_D5 + LCD_D6 + LCD_D7;
	PIO_SODR = (dat & 0x80) ? LCD_D7 : 0;
	PIO_SODR = (dat & 0x40) ? LCD_D6 : 0;
	PIO_SODR = (dat & 0x20) ? LCD_D5 : 0;
	PIO_SODR = (dat & 0x10) ? LCD_D4 : 0;
	wait_(83);	// wait_ 83us
	PIO_CODR = LCD_E;
	wait_(83);	// wait_ 83us
	PIO_SODR = LCD_E;
	PIO_CODR = LCD_D4 + LCD_D5 + LCD_D6 + LCD_D7;
	PIO_SODR = (dat & 0x8) ? LCD_D7 : 0;
	PIO_SODR = (dat & 0x4) ? LCD_D6 : 0;
	PIO_SODR = (dat & 0x2) ? LCD_D5 : 0;
	PIO_SODR = (dat & 0x1) ? LCD_D4 : 0;
	wait_(83);	// wait_ 83us
	PIO_CODR = LCD_E;
	wait_(83);	// wait_ 83us
	if(rs == 0) wait_(4000);	// wait_ 4ms
	PIO_CODR = LCD_RS;
#endif
#if defined(__H8300H__) || defined(__sh__)
	d = (rs == 0) ? 0 : LCD_RS;
	d |= LCD_E;
	d |= (dat & 0x80) ? LCD_D7 : 0;
	d |= (dat & 0x40) ? LCD_D6 : 0;
	d |= (dat & 0x20) ? LCD_D5 : 0;
	d |= (dat & 0x10) ? LCD_D4 : 0;
	*(lcd[minor].dr) = d;
	wait_(83);	// wait_ 83us
	d &= ~LCD_E;
	*(lcd[minor].dr) = d;
	wait_(83);	// wait_ 83us
	d = (rs == 0) ? 0 : LCD_RS;
	d |= LCD_E;
	d |= (dat & 0x8) ? LCD_D7 : 0;
	d |= (dat & 0x4) ? LCD_D6 : 0;
	d |= (dat & 0x2) ? LCD_D5 : 0;
	d |= (dat & 0x1) ? LCD_D4 : 0;
	*(lcd[minor].dr) = d;
	wait_(83);	// wait_ 83us
	d &= ~LCD_E;
	*(lcd[minor].dr) = d;
	wait_(83);	// wait_ 83us
	if(rs == 0) wait_(4000);	// wait_ 4ms
	d &= ~LCD_RS;
	*(lcd[minor].dr) = d;
#endif
}

static void lcd_clear(int minor) {
	lcd_out4(minor, 0, 0x01);
	lcd[minor].cur = 0;
}

static void lcd_putch(int minor, int _c) {
	int	c;

	c = _c & 0xff;
	switch(c) {
	case 0xfe:
		lcd_clear(minor);
		return;
	case '\r':
		lcd[minor].cur += lcd[minor].width;
		lcd[minor].cur = (lcd[minor].cur / lcd[minor].width) * lcd[minor].width;
		return;
	case '\n':
	case 0x8e:
		return;
	case 0x7e:
		c = 0xf3;
		break;
	}
	if(lcd[minor].cur >= lcd[minor].width * lcd[minor].height) {
		lcd_clear(minor);
	} else if(lcd[minor].cur == lcd[minor].width * 3) {
		lcd_out4(minor, 0, 0x80 + 0x54);
	} else if(lcd[minor].cur == lcd[minor].width * 2) {
		lcd_out4(minor, 0, 0x80 + 0x14);
	} else if(lcd[minor].cur == lcd[minor].width) {
		lcd_out4(minor, 0, 0x80 + 0x40);
	}
	lcd_out4(minor, 1, c);
	lcd[minor].cur++;
}

static int open_lcd(int minor, int option) {
	if(minor >= LCDNUM) return -1;
	lcd[minor].cur = 0;
#if defined(__arm__)
	PIO_PER = LCD_D4 + LCD_D5 + LCD_D6 + LCD_D7 + LCD_RS + LCD_E;
	PIO_OER = LCD_D4 + LCD_D5 + LCD_D6 + LCD_D7 + LCD_RS + LCD_E;
#endif
#if defined(__H8300H__) || defined(__sh__)
	lcd[minor].dr = get_dr(lcd[minor].port);
	ddr_set(lcd[minor].port, LCD_D4 + LCD_D5 + LCD_D6 + LCD_D7 + LCD_RS + LCD_E);
	*(lcd[minor].dr) = 0;
#endif
	wait_(4000);
	lcd_out8(minor, 0x23);
	lcd_out8(minor, 0x23);
	lcd_out8(minor, 0x23);
	lcd_out8(minor, 0x22);
	
	lcd_out4(minor, 0, 0x28);	// Function set
	lcd_out4(minor, 0, 0x0c);	// Display on
	lcd_out4(minor, 0, 0x06);	// Entry mode
	lcd_clear(minor);
	return 0;
}

static int close_lcd(int minor) {
	if(minor >= LCDNUM) return -1;
	return 0;
}

static int write_lcd(int minor, char *buffer, int size) {
	int	i;

	if(minor >= LCDNUM) return -1;
	for(i = 0;i < size;i++) lcd_putch(minor, buffer[i]);
	return size;	
}

static int read_lcd(int minor, char *buffer, int size) {
	if(minor >= LCDNUM) return -1;
	return 0;
}

static int seek_lcd(int minor, int pos) {
	if(minor >= LCDNUM) return -1;
	if(pos >= lcd[minor].width * lcd[minor].height) {
		return -1;
	} else if(pos >= lcd[minor].width * 3) {
		lcd_out4(minor, 0, 0x80 + 0x54 + pos - lcd[minor].width * 3);
	} else if(pos >= lcd[minor].width * 2) {
		lcd_out4(minor, 0, 0x80 + 0x14 + pos - lcd[minor].width * 2);
	} else if(pos >= lcd[minor].width) {
		lcd_out4(minor, 0, 0x80 + 0x40 + pos - lcd[minor].width);
	} else {
		lcd_out4(minor, 0, 0x80 + pos);
	}
	lcd[minor].cur = pos;
	return 0;
}

static int ioctl_lcd(int minor, int data, int op) {
	int	ret;

	if(minor >= LCDNUM) return -1;
	ret = 0;
	switch(op) {
	case DEV_INFO:
		break;
	default:
		ret = -1;
	}
	return ret;
}

static Functions func;

int device_main(int argc, char **argv) {
	int	minor;

#ifndef _KERNEL_
	int	i; /* ----- Append 2010/01/26 A.Ohnishi ----- */
	minor = 0;
	for(i = 1;i < argc;i++) {
		if(memcmp(argv[i], "lcd", 3) == 0) {
			minor = stoi(&argv[i][3]);
			if(minor >= LCDNUM) minor = 0;
		} else if(memcmp(argv[i], "p=", 2) == 0) {
			lcd[minor].port = stoi(&argv[i][2]);
		} else if(memcmp(argv[i], "d4=", 3) == 0) {
			LCD_D4 = 1 << stoi(&argv[i][3]);
		} else if(memcmp(argv[i], "d5=", 3) == 0) {
			LCD_D5 = 1 << stoi(&argv[i][3]);
		} else if(memcmp(argv[i], "d6=", 3) == 0) {
			LCD_D6 = 1 << stoi(&argv[i][3]);
		} else if(memcmp(argv[i], "d7=", 3) == 0) {
			LCD_D7 = 1 << stoi(&argv[i][3]);
		} else if(memcmp(argv[i], "rs=", 3) == 0) {
			LCD_RS = 1 << stoi(&argv[i][3]);
		} else if(memcmp(argv[i], "e=", 2) == 0) {
			LCD_E = 1 << stoi(&argv[i][2]);
		} else if(memcmp(argv[i], "row=", 4) == 0) {
			lcd[minor].height = stoi(&argv[i][4]);
		} else if(memcmp(argv[i], "col=", 4) == 0) {
			lcd[minor].width = stoi(&argv[i][4]);
		}
	}
#endif
	for(minor = 0;minor < LCDNUM;minor++) {
#ifdef _KERNEL_
  #if defined(__H8300H__) || defined(__sh__)
		lcd[minor].port = lcd_config[minor].port;
		LCD_D4 = 1 << lcd_config[minor].bits[0];
		LCD_D5 = 1 << lcd_config[minor].bits[1];
		LCD_D6 = 1 << lcd_config[minor].bits[2];
		LCD_D7 = 1 << lcd_config[minor].bits[3];
		LCD_RS = 1 << lcd_config[minor].bits[4];
		LCD_E = 1 << lcd_config[minor].bits[5];
  #endif
		lcd[minor].height = lcd_config[minor].bits[6];
		lcd[minor].width = lcd_config[minor].bits[7];
#endif
	}
	func.open_dev = open_lcd;
	func.close_dev = close_lcd;
	func.write_dev = write_lcd;
	func.read_dev = read_lcd;
	func.seek_dev = seek_lcd;
	func.ioctl_dev = ioctl_lcd;
	func.poll_dev = 0;
	strcpy(&(func.name[0]), "lcd");
	device_return(&func);
}
